import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from torch_activation import register_activation

@register_activation
class Sigmoid(nn.Module):
    r"""
    Applies the Sigmoid activation function:

    :math:`\text{Sigmoid}(z) = \frac{1}{1 + \exp(-z)}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input. 

    Examples::

        >>> m = nn.Sigmoid()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.Sigmoid(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(Sigmoid, self).__init__()

        self.inplace = inplace

    def forward(self, z) -> Tensor:
        if self.inplace:
            z.sigmoid_()
            return z
        else:
            return torch.sigmoid(z)
        
@register_activation
class Tanh(nn.Module):
    r"""
    Applies the Tanh activation function:

    :math:`\text{Tanh}(z) = \tanh(z)`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.Tanh()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.Tanh(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace: bool = False):
        super(Tanh, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        if self.inplace:
            z.tanh_()
            return z
        else:
            return torch.tanh(z)
        
        

@register_activation
class ShiftedScaledSigmoid(nn.Module):
    r"""
    Applies the Shifted Scaled Sigmoid activation function:

    :math:`\text{ShiftedScaledSigmoid}(z) = \frac{1}{1 + \exp(-a(z-b))}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Shift parameter. Default: 0.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ShiftedScaledSigmoid(a=2.0, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.ShiftedScaledSigmoid(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 0.0, inplace: bool = False):
        super(ShiftedScaledSigmoid, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor([b]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        return torch.sigmoid(self.a * (z - self.b))

    def _forward_inplace(self, z):
        z.sub_(self.b).mul_(self.a)
        z.sigmoid_()
        return z


@register_activation
class VariantSigmoidFunction(nn.Module):
    r"""
    Applies the Variant Sigmoid Function activation:

    :math:`\text{VariantSigmoidFunction}(z) = \frac{a}{1 + \exp(-bz)} - c`

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Slope parameter. Default: 1.0
        c (float, optional): Offset parameter. Default: 0.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.VariantSigmoidFunction(a=2.0, b=1.5, c=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.VariantSigmoidFunction(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(
        self, a: float = 1.0, b: float = 1.0, c: float = 0.0, inplace: bool = False
    ):
        super(VariantSigmoidFunction, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor([b]))
        self.c = nn.Parameter(torch.tensor([c]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        return self.a * torch.sigmoid(self.b * z) - self.c

    def _forward_inplace(self, z):
        z.mul_(self.b).sigmoid_().mul_(self.a).sub_(self.c)
        return z


@register_activation
class STanh(nn.Module):
    r"""
    Applies the Scaled Hyperbolic Tangent activation function:

    :math:`\text{STanh}(z) = a \tanh(bz)`

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Slope parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.STanh(a=1.7, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.STanh(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0, inplace: bool = False):
        super(STanh, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor([b]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        return self.a * torch.tanh(self.b * z)

    def _forward_inplace(self, z):
        z.mul_(self.b).tanh_().mul_(self.a)
        return z

# FIXME: Revise this
# @register_activation
# class BiModalDerivativeSigmoid(nn.Module):
#     r"""
#     Applies the Bi-Modal Derivative Sigmoid activation function:

#     :math:`\text{BiModalDerivativeSigmoid}(z) = \frac{a}{1 + \exp(-bz)} - \frac{1}{2} \left( \frac{1}{1 + \exp(-z)} + \frac{1}{1 + \exp(-z-b)} \right)`

#     Args:
#         b (float, optional): Shift parameter. Default: 1.0

#     Shape:
#         - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
#         - Output: :math:`(*)`, same shape as the input.

#     Examples::

#         >>> m = nn.BiModalDerivativeSigmoid(b=2.0)
#         >>> x = torch.randn(2)
#         >>> output = m(x)
#     """

#     def __init__(self, b: float = 1.0):
#         super(BiModalDerivativeSigmoid, self).__init__()
#         self.b = nn.Parameter(torch.tensor([b]))

#     def forward(self, z):
#         first_term = self.a * torch.sigmoid(self.b * z)
#         second_term = torch.sigmoid(z) + torch.sigmoid(z + self.b)
#         return 0.5 * (first_term - second_term)


@register_activation
class Arctan(nn.Module):
    r"""
    Applies the Arctan activation function:

    :math:`\text{Arctan}(z) = \arctan(z)`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.Arctan()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Arctan, self).__init__()

    def forward(self, z) -> Tensor:
        return torch.atan(z)


@register_activation
class ArctanGR(nn.Module):
    r"""
    Applies the ArctanGR activation function:

    :math:`\text{ArctanGR}(z) = \frac{\arctan(z)}{1 + \sqrt{2}}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ArctanGR()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ArctanGR, self).__init__()
        self.scale_factor = 1.0 / (1.0 + torch.sqrt(torch.tensor(2.0)))

    def forward(self, z) -> Tensor:
        return torch.atan(z) * self.scale_factor

@register_activation
class SigmoidAlgebraic(nn.Module):
    r"""
    Applies the Sigmoid Algebraic activation function:

    :math:`\text{SigmoidAlgebraic}(z) = \frac{1}{1 + \exp\left(-\frac{z(1 + a|z|)}{1 + |z|(1 + a|z|)}\right)}`

    Args:
        a (float, optional): Shape parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SigmoidAlgebraic(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SigmoidAlgebraic(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super(SigmoidAlgebraic, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        abs_z = torch.abs(z)
        a_abs_z = self.a * abs_z
        numerator = z * (1 + a_abs_z)
        denominator = 1 + abs_z * (1 + a_abs_z)
        return torch.sigmoid(numerator / denominator)

    def _forward_inplace(self, z):
        # Cannot be done fully in-place due to the complex calculation
        abs_z = torch.abs(z)
        a_abs_z = self.a * abs_z
        numerator = z.clone() * (1 + a_abs_z)
        denominator = 1 + abs_z * (1 + a_abs_z)
        z.copy_(numerator / denominator).sigmoid_()
        return z


@register_activation
class TripleStateSigmoid(nn.Module):
    r"""
    Applies the Triple State Sigmoid activation function:

    :math:`\text{TripleStateSigmoid}(z) = \frac{1}{1 + \exp(-z)} + \frac{1}{1 + \exp(-z+a)} + \frac{1}{1 + \exp(-z+b)}`

    Args:
        a (float, optional): First shift parameter. Default: 1.0
        b (float, optional): Second shift parameter. Default: 2.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.TripleStateSigmoid(a=1.5, b=2.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.TripleStateSigmoid(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 2.0, inplace: bool = False):
        super(TripleStateSigmoid, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor([b]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        return torch.sigmoid(z) + torch.sigmoid(z - self.a) + torch.sigmoid(z - self.b)

    def _forward_inplace(self, z):
        # Cannot be done fully in-place due to the need for the original z value
        result = torch.sigmoid(z.clone())
        result.add_(torch.sigmoid(z - self.a))
        result.add_(torch.sigmoid(z - self.b))
        z.copy_(result)
        return z


@register_activation
class ImprovedLogisticSigmoid(nn.Module):
    r"""
    Applies the Improved Logistic Sigmoid activation function:

    :math:`\text{ImprovedLogisticSigmoid}(z) = \begin{cases} 
    a(z-b) + \sigma(b), & z \geq b \\ 
    \sigma(z), & -b < z < b \\ 
    a(z+b) + \sigma(-b), & z \leq -b 
    \end{cases}`

    Args:
        a (float, optional): Slope parameter. Default: 0.2
        b (float, optional): Threshold parameter. Default: 2.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ImprovedLogisticSigmoid(a=0.1, b=3.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.ImprovedLogisticSigmoid(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.2, b: float = 2.0, inplace: bool = False):
        super(ImprovedLogisticSigmoid, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor([b]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        sig_b = torch.sigmoid(self.b)
        sig_neg_b = torch.sigmoid(-self.b)

        upper_region = self.a * (z - self.b) + sig_b
        middle_region = torch.sigmoid(z)
        lower_region = self.a * (z + self.b) + sig_neg_b

        result = torch.where(z >= self.b, upper_region, middle_region)
        result = torch.where(z <= -self.b, lower_region, result)

        return result

    def _forward_inplace(self, z):
        # Cannot be done fully in-place due to the conditional nature
        sig_b = torch.sigmoid(self.b)
        sig_neg_b = torch.sigmoid(-self.b)

        upper_mask = z >= self.b
        lower_mask = z <= -self.b
        middle_mask = ~(upper_mask | lower_mask)

        result = z.clone()

        # Apply each region's transformation
        result[upper_mask] = self.a * (z[upper_mask] - self.b) + sig_b
        result[middle_mask] = torch.sigmoid(z[middle_mask])
        result[lower_mask] = self.a * (z[lower_mask] + self.b) + sig_neg_b

        z.copy_(result)
        return z


@register_activation
class SigLin(nn.Module):
    r"""
    Applies the SigLin activation function:

    :math:`\text{SigLin}(z) = \sigma(z) + az`

    Args:
        a (float, optional): Linear coefficient. Default: 0.2
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SigLin(a=0.1)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SigLin(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 0.2, inplace: bool = False):
        super(SigLin, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        return torch.sigmoid(z) + self.a * z

    def _forward_inplace(self, z):
        # Cannot be done fully in-place
        z_copy = z.clone()
        z.sigmoid_().add_(self.a * z_copy)
        return z


@register_activation
class PenalizedHyperbolicTangent(nn.Module):
    r"""
    Applies the Penalized Hyperbolic Tangent activation function:

    :math:`\text{PenalizedHyperbolicTangent}(z) = \begin{cases} 
    \tanh(z), & z \geq 0 \\ 
    \frac{\tanh(z)}{a}, & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Penalty factor for negative inputs. Default: 2.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.PenalizedHyperbolicTangent(a=3.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.PenalizedHyperbolicTangent(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 2.0, inplace: bool = False):
        super(PenalizedHyperbolicTangent, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        tanh_z = torch.tanh(z)
        return torch.where(z >= 0, tanh_z, tanh_z / self.a)

    def _forward_inplace(self, z):
        # Cannot be done fully in-place due to the conditional nature
        neg_mask = z < 0
        z.tanh_()
        z[neg_mask].div_(self.a)
        return z


@register_activation
class SoftRootSign(nn.Module):
    r"""
    Applies the Soft Root Sign activation function:

    :math:`\text{SoftRootSign}(z) = \frac{z}{\sqrt[a]{1 + \exp\left(-\frac{z}{b}\right)}}`

    Args:
        a (float, optional): Root parameter. Default: 2.0
        b (float, optional): Scale parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SoftRootSign(a=3.0, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SoftRootSign(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 2.0, b: float = 1.0, inplace: bool = False):
        super(SoftRootSign, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor([b]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        denominator = torch.pow(1 + torch.exp(-z / self.b), 1 / self.a)
        return z / denominator

    def _forward_inplace(self, z):
        # Cannot be done fully in-place
        z_copy = z.clone()
        denominator = torch.pow(1 + torch.exp(-z / self.b), 1 / self.a)
        z.copy_(z_copy / denominator)
        return z


@register_activation
class SoftClipping(nn.Module):
    r"""
    Applies the Soft Clipping activation function:

    :math:`\text{SoftClipping}(z) = \frac{1}{a} \ln\left(\frac{1 + \exp(az)}{1 + \exp(a(z-1))}\right)`

    Args:
        a (float, optional): Sharpness parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SoftClipping(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SoftClipping(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, inplace: bool = False):
        super(SoftClipping, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        numerator = 1 + torch.exp(self.a * z)
        denominator = 1 + torch.exp(self.a * (z - 1))
        return torch.log(numerator / denominator) / self.a

    def _forward_inplace(self, z):
        # Cannot be done fully in-place
        numerator = 1 + torch.exp(self.a * z)
        denominator = 1 + torch.exp(self.a * (z - 1))
        z.copy_(torch.log(numerator / denominator) / self.a)
        return z


@register_activation
class Hexpo(nn.Module):
    r"""
    Applies the Hexpo activation function:

    :math:`\text{Hexpo}(z) = \begin{cases} 
    -a \exp\left(-\frac{z}{b}\right) - 1, & z \geq 0 \\ 
    c \exp\left(-\frac{z}{d}\right) - 1, & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Positive scale parameter. Default: 1.0
        b (float, optional): Positive decay parameter. Default: 1.0
        c (float, optional): Negative scale parameter. Default: 1.0
        d (float, optional): Negative decay parameter. Default: 1.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.Hexpo(a=1.5, b=0.5, c=2.0, d=0.7)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.Hexpo(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(
        self,
        a: float = 1.0,
        b: float = 1.0,
        c: float = 1.0,
        d: float = 1.0,
        inplace: bool = False,
    ):
        super(Hexpo, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor([b]))
        self.c = nn.Parameter(torch.tensor([c]))
        self.d = nn.Parameter(torch.tensor([d]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        pos_part = -self.a * torch.exp(-z / self.b) - 1
        neg_part = self.c * torch.exp(-z / self.d) - 1
        return torch.where(z >= 0, pos_part, neg_part)

    def _forward_inplace(self, z):
        # Cannot be done fully in-place due to the conditional nature
        pos_mask = z >= 0
        neg_mask = ~pos_mask

        result = z.clone()
        result[pos_mask] = -self.a * torch.exp(-z[pos_mask] / self.b) - 1
        result[neg_mask] = self.c * torch.exp(-z[neg_mask] / self.d) - 1

        z.copy_(result)
        return z


@register_activation
class Softsign(nn.Module):
    r"""
    Applies the Softsign activation function:

    :math:`\text{Softsign}(z) = \frac{z}{1 + |z|}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.Softsign()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Softsign, self).__init__()

    def forward(self, z) -> Tensor:
        return z / (1 + torch.abs(z))


@register_activation
class SmoothStep(nn.Module):
    r"""
    Applies the Smooth Step activation function:

    :math:`\text{SmoothStep}(z) = \begin{cases} 
    1, & z \geq \frac{a}{2} \\ 
    \frac{2}{a^3} z^3 - \frac{3}{2a} z + \frac{1}{2}, & -\frac{a}{2} \leq z \leq \frac{a}{2} \\ 
    0, & z \leq -\frac{a}{2} 
    \end{cases}`

    Args:
        a (float, optional): Width parameter. Default: 2.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SmoothStep(a=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.SmoothStep(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, a: float = 2.0, inplace: bool = False):
        super(SmoothStep, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return self._forward_inplace(z) if self.inplace else self._forward(z)

    def _forward(self, z):
        half_a = self.a / 2

        # Calculate the polynomial for the middle region
        cubic_term = (2 / (self.a**3)) * (z**3)
        linear_term = (3 / (2 * self.a)) * z
        constant_term = 0.5
        middle_region = cubic_term - linear_term + constant_term

        # Apply the piecewise function
        result = torch.ones_like(z)
        middle_mask = (z > -half_a) & (z < half_a)
        lower_mask = z <= -half_a

        result[middle_mask] = middle_region[middle_mask]
        result[lower_mask] = 0.0

        return result

    def _forward_inplace(self, z):
        # Cannot be done fully in-place due to the conditional nature
        half_a = self.a / 2

        # Create masks for each region
        upper_mask = z >= half_a
        middle_mask = (z > -half_a) & (z < half_a)
        lower_mask = z <= -half_a

        # Calculate the polynomial for the middle region
        middle_values = z[middle_mask]
        cubic_term = (2 / (self.a**3)) * (middle_values**3)
        linear_term = (3 / (2 * self.a)) * middle_values
        middle_result = cubic_term - linear_term + 0.5

        # Apply the piecewise function
        result = z.clone()
        result[upper_mask] = 1.0
        result[middle_mask] = middle_result
        result[lower_mask] = 0.0

        z.copy_(result)
        return z


@register_activation
class ElliottActivationFunction(nn.Module):
    r"""
    Applies the Elliott Activation Function:

    :math:`\text{ElliottActivationFunction}(z) = \frac{0.5z}{1 + |z|} + 0.5`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ElliottActivationFunction()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ElliottActivationFunction, self).__init__()

    def forward(self, z) -> Tensor:
        return (0.5 * z) / (1 + torch.abs(z)) + 0.5


@register_activation
class SincSigmoid(nn.Module):
    r"""
    Applies the Sinc Sigmoid activation function:

    :math:`\text{SincSigmoid}(z) = \text{sinc}(\sigma(z))`

    where :math:`\text{sinc}(x) = \frac{\sin(\pi x)}{\pi x}` if :math:`x \neq 0`, and 1 if :math:`x = 0`.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SincSigmoid()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(SincSigmoid, self).__init__()

    def forward(self, z) -> Tensor:
        sigmoid_z = torch.sigmoid(z)
        # Handle the case where sigmoid_z is 0 to avoid division by zero
        result = torch.ones_like(z)
        nonzero_mask = sigmoid_z != 0

        # Calculate sinc for non-zero values
        pi_sigmoid_z = torch.pi * sigmoid_z[nonzero_mask]
        result[nonzero_mask] = torch.sin(pi_sigmoid_z) / pi_sigmoid_z

        return result


@register_activation
class SigmoidGumbel(nn.Module):
    r"""
    Applies the Sigmoid Gumbel activation function:

    :math:`\text{SigmoidGumbel}(z) = \frac{1}{1 + \exp(-z) \exp(-\exp(-z))}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SigmoidGumbel()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(SigmoidGumbel, self).__init__()

    def forward(self, z) -> Tensor:
        neg_z = -z
        return 1 / (1 + torch.exp(neg_z) * torch.exp(-torch.exp(neg_z)))

@register_activation
class NewSigmoid(nn.Module):
    r"""
    Applies the New Sigmoid activation function:

    :math:`\text{NewSigmoid}(z) = \frac{\exp(z) - \exp(-z)}{2(\exp(2z) + \exp(-2z))}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.NewSigmoid()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(NewSigmoid, self).__init__()

    def forward(self, z) -> Tensor:
        exp_z = torch.exp(z)
        exp_neg_z = torch.exp(-z)
        exp_2z = torch.exp(2 * z)
        exp_neg_2z = torch.exp(-2 * z)

        numerator = exp_z - exp_neg_z
        denominator = 2 * (exp_2z + exp_neg_2z)

        return numerator / denominator


@register_activation
class Root2sigmoid(nn.Module):
    r"""
    Applies the Root2sigmoid activation function:

    :math:`\text{Root2sigmoid}(z) = \frac{\sqrt{2}z}{\sqrt{2^{-2z}} + \sqrt{2^{2z}}}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.Root2sigmoid()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Root2sigmoid, self).__init__()
        self.sqrt2 = torch.sqrt(torch.tensor(2.0))

    def forward(self, z) -> Tensor:
        sqrt2_z = self.sqrt2 * z
        neg_2z = -2 * z
        pos_2z = 2 * z

        denominator = torch.sqrt(torch.pow(2, neg_2z)) + torch.sqrt(
            torch.pow(2, pos_2z)
        )
        return sqrt2_z / denominator


@register_activation
class LogLog(nn.Module):
    r"""
    Applies the LogLog activation function:

    :math:`\text{LogLog}(z) = \exp(-\exp(-z))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.LogLog()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(LogLog, self).__init__()

    def forward(self, z) -> Tensor:
        return torch.exp(-torch.exp(-z))


@register_activation
class ComplementaryLogLog(nn.Module):
    r"""
    Applies the Complementary LogLog activation function:

    :math:`\text{ComplementaryLogLog}(z) = 1 - \exp(-\exp(-z))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ComplementaryLogLog()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ComplementaryLogLog, self).__init__()

    def forward(self, z) -> Tensor:
        return 1 - torch.exp(-torch.exp(-z))


@register_activation
class ModifiedComplementaryLogLog(nn.Module):
    r"""
    Applies the Modified Complementary LogLog activation function:

    :math:`\text{ModifiedComplementaryLogLog}(z) = 1 - 2\exp(-0.7\exp(-z))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ModifiedComplementaryLogLog()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ModifiedComplementaryLogLog, self).__init__()

    def forward(self, z) -> Tensor:
        return 1 - 2 * torch.exp(-0.7 * torch.exp(-z))


@register_activation
class SechSig(nn.Module):
    r"""
    Applies the SechSig activation function:

    :math:`\text{SechSig}(z) = (z + \text{sech}(z))\sigma(z)`

    where :math:`\text{sech}(z) = \frac{2}{e^z + e^{-z}}` is the hyperbolic secant.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SechSig()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(SechSig, self).__init__()

    def forward(self, z) -> Tensor:
        # Calculate hyperbolic secant: sech(z) = 2/(e^z + e^-z)
        exp_z = torch.exp(z)
        exp_neg_z = torch.exp(-z)
        sech_z = 2 / (exp_z + exp_neg_z)

        return (z + sech_z) * torch.sigmoid(z)


@register_activation
class ParametricSechSig(nn.Module):
    r"""
    Applies the Parametric SechSig activation function:

    :math:`\text{ParametricSechSig}(z) = (z + a\cdot \text{sech}(z+a))\sigma(z)`

    where :math:`\text{sech}(z) = \frac{2}{e^z + e^{-z}}` is the hyperbolic secant.

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ParametricSechSig(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(ParametricSechSig, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))

    def forward(self, z) -> Tensor:
        # Calculate shifted input
        z_plus_a = z + self.a

        # Calculate hyperbolic secant: sech(z+a) = 2/(e^(z+a) + e^-(z+a))
        exp_z_plus_a = torch.exp(z_plus_a)
        exp_neg_z_plus_a = torch.exp(-z_plus_a)
        sech_z_plus_a = 2 / (exp_z_plus_a + exp_neg_z_plus_a)

        return (z + self.a * sech_z_plus_a) * torch.sigmoid(z)


@register_activation
class TanhSig(nn.Module):
    r"""
    Applies the TanhSig activation function:

    :math:`\text{TanhSig}(z) = (z + \tanh(z))\sigma(z)`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.TanhSig()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(TanhSig, self).__init__()

    def forward(self, z) -> Tensor:
        return (z + torch.tanh(z)) * torch.sigmoid(z)


@register_activation
class ParametricTanhSig(nn.Module):
    r"""
    Applies the Parametric TanhSig activation function:

    :math:`\text{ParametricTanhSig}(z) = (z + a\cdot \tanh(z+a))\sigma(z)`

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.ParametricTanhSig(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(ParametricTanhSig, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))

    def forward(self, z) -> Tensor:
        return (z + self.a * torch.tanh(z + self.a)) * torch.sigmoid(z)


@register_activation
class MultistateActivationFunction(nn.Module):
    r"""
    Applies the Multistate Activation Function:

    :math:`\text{MultistateActivationFunction}(z) = a + \sum_{k=1}^N \frac{1}{1 + \exp(-z+b_k)}`

    Args:
        a (float, optional): Offset parameter. Default: 0.0
        b (list of float, optional): List of shift parameters. Default: [1.0, 2.0]

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.MultistateActivationFunction(a=0.5, b=[0.5, 1.5, 2.5])
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 0.0, b: list = None):
        super(MultistateActivationFunction, self).__init__()
        if b is None:
            b = [1.0, 2.0]
        self.a = nn.Parameter(torch.tensor([a]))
        self.b = nn.Parameter(torch.tensor(b))

    def forward(self, z) -> Tensor:
        result = self.a.clone()

        for k in range(len(self.b)):
            result = result + torch.sigmoid(z - self.b[k])

        return result


@register_activation
class SymmetricalMSAF(nn.Module):
    r"""
    Applies the Symmetrical Multistate Activation Function:

    :math:`\text{SymmetricalMSAF}(z) = -1 + \frac{1}{1 + \exp(-z)} + \frac{1}{1 + \exp(-z-a)}`

    Args:
        a (float, optional): Shift parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SymmetricalMSAF(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(SymmetricalMSAF, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))

    def forward(self, z) -> Tensor:
        return -1 + torch.sigmoid(z) + torch.sigmoid(z - self.a)


@register_activation
class Rootsig(nn.Module):
    r"""
    Applies the Rootsig activation function:

    :math:`\text{Rootsig}(z) = \frac{az}{\sqrt{1 + a^2z^2}}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.Rootsig(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(Rootsig, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))

    def forward(self, z) -> Tensor:
        a_z = self.a * z
        return a_z / torch.sqrt(1 + a_z * a_z)


@register_activation
class UnnamedSigmoid1(nn.Module):
    # TODO: Naming
    r"""
    :note: The name "UnnamesSigmoid1" derived from "3.2.25 Rootsig and others" entry, particularly the first from from "others" part. Leave it here until I find a better name.
    Applies the Unnamed Sigmoid 1 activation function:

    :math:`\text{UnnamedSigmoid1}(z) = z \cdot \text{sgn}(z) \sqrt{z^{-a} - a^{-2}}`

    Args:
        a (float, optional): Shape parameter. Default: 2.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.UnnamedSigmoid1(a=3.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 2.0):
        super(UnnamedSigmoid1, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))
        self.a_squared_inv = 1 / (a * a)

    def forward(self, z) -> Tensor:
        # Handle the sign of z
        sign_z = torch.sign(z)
        abs_z = torch.abs(z)

        # Avoid division by zero or negative values under sqrt
        eps = 1e-6
        safe_z = torch.clamp(abs_z, min=eps)

        # Calculate z^(-a)
        z_pow_neg_a = torch.pow(safe_z, -self.a)

        # Calculate sqrt term, ensuring the value under sqrt is positive
        sqrt_term = torch.sqrt(torch.clamp(z_pow_neg_a - self.a_squared_inv, min=0))

        return z * sign_z * sqrt_term


@register_activation
class UnnamedSigmoid2(nn.Module):
    # TODO: Naming
    r"""
    :note: The name "UnnamesSigmoid2" derived from "3.2.25 Rootsig and others" entry, particularly the second from from "others" part. Leave it here until I find a better name.
    Applies the Unnamed Sigmoid 2 activation function:

    :math:`\text{UnnamedSigmoid2}(z) = \frac{az}{1 + |az|}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.UnnamedSigmoid2(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(UnnamedSigmoid2, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))

    def forward(self, z) -> Tensor:
        a_z = self.a * z
        return a_z / (1 + torch.abs(a_z))


@register_activation
class UnnamedSigmoid3(nn.Module):
    # TODO: Naming
    r"""
    :note: The name "UnnamedSigmoid3" derived from "3.2.25 Rootsig and others" entry, particularly the third from from "others" part. Leave it here until I find a better name.
    Applies the Unnamed Sigmoid 3 activation function:

    :math:`\text{UnnamedSigmoid3}(z) = \frac{az}{\sqrt{1 + a^2z^2}}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.UnnamedSigmoid3(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(UnnamedSigmoid3, self).__init__()
        self.a = nn.Parameter(torch.tensor([a]))

    def forward(self, z) -> Tensor:
        a_z = self.a * z
        return a_z / torch.sqrt(1 + a_z * a_z)


@register_activation
class SigmoidTanhCombinations(nn.Module):
    r"""
    Applies the Sigmoid-Tanh Combinations activation function:

    :math:`\text{SigmoidTanhCombinations}(z) = \begin{cases} 
    g(z), & z \geq 0 \\ 
    h(z), & z < 0 
    \end{cases}`

    where g(z) and h(z) are user-defined functions, defaulting to sigmoid and tanh respectively.

    Args:
        g_func (callable, optional): Function to use for positive inputs. Default: torch.sigmoid
        h_func (callable, optional): Function to use for negative inputs. Default: torch.tanh

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = nn.SigmoidTanhCombinations()
        >>> x = torch.randn(2)
        >>> output = m(x)
        
        >>> # Custom functions
        >>> import torch.nn.functional as F
        >>> m = nn.SigmoidTanhCombinations(g_func=F.relu, h_func=torch.sigmoid)
        >>> output = m(x)
    """

    def __init__(self, g_func=torch.sigmoid, h_func=torch.tanh):
        super(SigmoidTanhCombinations, self).__init__()
        self.g_func = g_func
        self.h_func = h_func

    def forward(self, z) -> Tensor:
        pos_mask = z >= 0
        neg_mask = ~pos_mask

        result = torch.zeros_like(z)
        result[pos_mask] = self.g_func(z[pos_mask])
        result[neg_mask] = self.h_func(z[neg_mask])

        return result



import torch
import torch.nn as nn
from torch import Tensor
import math

from torch_activation import register_activation

@register_activation
class Binary(nn.Module):
    r"""
    Applies the Binary activation function:

    :math:`\text{Binary}(z) = \begin{cases} 
    0, & z < 0 \\
    1, & z \geq 0 
    \end{cases}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.
    """

    def __init__(self, inplace: bool = False):
        super(Binary, self).__init__()
        self.inplace = inplace

    def forward(self, z) -> Tensor:
        return _Binary.apply(z)


class _Binary(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input):
        ctx.save_for_backward(input)
        return (input >= 0).float()

    @staticmethod
    def backward(ctx, grad_output):
        # Straight-through estimator
        # Pass the gradient through unchanged
        return grad_output


@register_activation
class Sine(nn.Module):
    r"""
    Applies the Sine activation function:

    :math:`\text{Sine}(z) = \sin(\pi \cdot z)`

    Args:
        omega (float, optional): frequency of the sine wave. Default: ``math.pi``
        inplace (bool, optional): parameter kept for API consistency, but sine operation 
                                 cannot be done in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.
    """

    def __init__(self, omega: float = math.pi, inplace: bool = False):
        super(Sine, self).__init__()
        self.omega = omega
        self.inplace = inplace  # Unused

    def forward(self, z) -> Tensor:
        return torch.sin(self.omega * z)


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from torch_activation import register_activation

@register_activation
class Softmax(nn.Module):
    r"""
    Applies the Softmax function:

    :math:`\text{Softmax}(z_j) = \frac{\exp(z_j)}{\sum_{k=1}^{N} \exp(z_k)}`

    where :math:`z_j` is the input of neuron j in a softmax layer consisting of N neurons.

    Args:
        dim (int, optional): A dimension along which Softmax will be computed. Default: -1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Softmax(dim=1)
        >>> input = torch.randn(2, 3)
        >>> output = m(input)
    """
    def __init__(self, dim=-1, inplace=False):
        super(Softmax, self).__init__()
        self.dim = dim
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            x.softmax_(dim=self.dim)
            return x
        else:
            return F.softmax(x, dim=self.dim)


@register_activation
class BetaSoftmax(nn.Module):
    r"""
    Applies the β-Softmax function:

    :math:`\text{β-Softmax}(z_j) = \frac{\exp(β \cdot z_j)}{\sum_{k=1}^{N} \exp(β \cdot z_k)}`

    where :math:`z_j` is the input of neuron j in a softmax layer consisting of N neurons,
    and :math:`β` is a trainable or fixed parameter controlling the sharpness of the distribution.

    Args:
        beta (float, optional): Initial value for the beta parameter. Default: 1.0
        trainable (bool, optional): If True, beta is a trainable parameter. Default: False
        dim (int, optional): A dimension along which Softmax will be computed. Default: -1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = BetaSoftmax(beta=2.0, trainable=True)
        >>> input = torch.randn(2, 3)
        >>> output = m(input)
    """
    def __init__(self, beta=1.0, trainable=False, dim=-1, inplace=False):
        super(BetaSoftmax, self).__init__()
        self.dim = dim
        self.inplace = inplace
        
        if trainable:
            self.beta = nn.Parameter(torch.tensor(beta))
        else:
            self.register_buffer('beta', torch.tensor(beta))

    def forward(self, x: Tensor) -> Tensor:
        if self.inplace:
            x = x.mul_(self.beta)
            x.softmax_(dim=self.dim)
            return x
        else:
            scaled_x = self.beta * x
            return F.softmax(scaled_x, dim=self.dim)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
import math

from torch_activation import register_activation


@register_activation
class Swish(nn.Module):
    r"""
    Applies the Swish activation function:

    :math:`\text{Swish}(x) = x \cdot \sigma(a \cdot x)`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 1.0
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.Swish(a=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.Swish(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        result = x * torch.sigmoid(self.a * x)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class AHAF(nn.Module):
    r"""
    Applies the Adaptive Hybrid Activation Function:

    :math:`\text{AHAF}(x) = a \cdot x \cdot \sigma(b \cdot x)`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Scaling parameter. Default: 1.0
        b (float, optional): Parameter controlling the shape of the function. Default: 1.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.AHAF(a=1.0, b=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.AHAF(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        result = self.a * x * torch.sigmoid(self.b * x)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class PSSiLU(nn.Module):
    r"""
    Applies the Parametric Shifted SiLU function:

    :math:`\text{PSSiLU}(x) = x \cdot \frac{\sigma(a \cdot x) - b}{1 - b}`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 1.0
        b (float, optional): Shift parameter. Default: 0.5
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.PSSiLU(a=1.0, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PSSiLU(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 0.5, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            # Ensure b is less than 1 to avoid division by zero
            self.b = nn.Parameter(Tensor([min(b, 0.99)]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([min(b, 0.99)])  # Ensure b is less than 1

    def forward(self, x) -> Tensor:
        # Compute the shifted and normalized sigmoid
        shifted_sigmoid = (torch.sigmoid(self.a * x) - self.b) / (1 - self.b)
        result = x * shifted_sigmoid
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class ESwish(nn.Module):
    r"""
    Applies the E-Swish activation function:

    :math:`\text{E-swish}(x) = a \cdot x \cdot \sigma(x)`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Scaling parameter, recommended in range [1, 2]. Default: 1.5
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.ESwish(a=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ESwish(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.5, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        result = self.a * x * torch.sigmoid(x)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class ACONB(nn.Module):
    r"""
    Applies the ACON-B activation function:

    :math:`\text{ACON-B}(x) = (1 - b) \cdot x \cdot \sigma(a \cdot (1 - b) \cdot x) + b \cdot x`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 1.0
        b (float, optional): Parameter controlling the linear component. Default: 0.25
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.ACONB(a=1.0, b=0.25)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ACONB(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 0.25, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            # Ensure b is between 0 and 1
            self.b = nn.Parameter(Tensor([max(0.0, min(b, 1.0))]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([max(0.0, min(b, 1.0))])  # Ensure b is between 0 and 1

    def forward(self, x) -> Tensor:
        one_minus_b = 1 - self.b
        swish_part = one_minus_b * x * torch.sigmoid(self.a * one_minus_b * x)
        linear_part = self.b * x
        result = swish_part + linear_part
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class ACONC(nn.Module):
    r"""
    Applies the ACON-C activation function:

    :math:`\text{ACON-C}(x) = (c - b) \cdot x \cdot \sigma(a \cdot (c - b) \cdot x) + b \cdot x`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 1.0
        b (float, optional): Parameter controlling the linear component. Default: 0.0
        c (float, optional): Parameter controlling the swish component. Default: 1.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.ACONC(a=1.0, b=0.0, c=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ACONC(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 0.0, 
        c: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
            self.c = nn.Parameter(Tensor([c]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])
            self.c = Tensor([c])

    def forward(self, x) -> Tensor:
        c_minus_b = self.c - self.b
        swish_part = c_minus_b * x * torch.sigmoid(self.a * c_minus_b * x)
        linear_part = self.b * x
        result = swish_part + linear_part
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class PSGU(nn.Module):
    r"""
    Applies the Parameterized Self-Circulating Gating Unit function:

    :math:`\text{PSGU}(x) = x \cdot \tanh(a \cdot \sigma(x))`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 0.5
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.PSGU(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PSGU(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 0.5, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        result = x * torch.tanh(self.a * torch.sigmoid(x))
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class TBSReLUl(nn.Module):
    r"""
    Applies the Tangent-Bipolar-Sigmoid ReLU Learnable function:

    :math:`\text{TBSReLUl}(x) = x \cdot \tanh\left(a \cdot \frac{1 - \exp(-x)}{1 + \exp(-x)}\right)`

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 0.5
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.TBSReLUl(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.TBSReLUl(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 0.5, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        # Calculate bipolar sigmoid: (1 - exp(-x)) / (1 + exp(-x))
        bipolar_sigmoid = (1 - torch.exp(-x)) / (1 + torch.exp(-x))
        result = x * torch.tanh(self.a * bipolar_sigmoid)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class PATS(nn.Module):
    r"""
    Applies the PATS activation function:

    :math:`\text{PATS}(x) = x \cdot \arctan(a \cdot \pi \cdot \sigma(x))`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 0.625
        lower_bound (float, optional): Lower bound for sampling a. Default: 0.5
        upper_bound (float, optional): Upper bound for sampling a. Default: 0.75
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.PATS(a=0.625)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PATS(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 0.625, 
        lower_bound: float = 0.5,
        upper_bound: float = 0.75,
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        
        if learnable:
            # Initialize with a value in the valid range
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        # If not in training mode or not learnable, use the fixed parameter
        if not self.training or not isinstance(self.a, nn.Parameter):
            a_value = self.a
        else:
            # During training with learnable parameter, sample from uniform distribution
            a_value = torch.rand_like(self.a) * (self.upper_bound - self.lower_bound) + self.lower_bound
            
        result = x * torch.arctan(a_value * math.pi * torch.sigmoid(x))
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class AQuLU(nn.Module):
    r"""
    Applies the Adaptive Quadratic Linear Unit function:

    :math:`\text{AQuLU}(x) = \begin{cases} 
        x, & x \geq \frac{1 - b}{a} \\
        a \cdot x^2 + b \cdot x, & -\frac{b}{a} \leq x < \frac{1 - b}{a} \\
        0, & x < -\frac{b}{a} 
    \end{cases}`

    Args:
        a (float, optional): Parameter controlling the quadratic component. Default: 0.2
        b (float, optional): Parameter controlling the linear component. Default: 0.1
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.AQuLU(a=0.2, b=0.1)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.AQuLU(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 0.2, 
        b: float = 0.1, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            # Ensure a is positive to avoid division by zero
            self.a = nn.Parameter(Tensor([max(1e-6, a)]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([max(1e-6, a)])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        # Calculate thresholds
        upper_threshold = (1 - self.b) / self.a
        lower_threshold = -self.b / self.a
        
        # Create masks for different regions
        mask_upper = x >= upper_threshold
        mask_middle = (x >= lower_threshold) & (x < upper_threshold)
        mask_lower = x < lower_threshold
        
        if self.inplace:
            # Create a copy to avoid modifying during computation
            result = x.clone()
            
            # Apply different functions to different regions
            result[mask_upper] = x[mask_upper]
            result[mask_middle] = self.a * x[mask_middle]**2 + self.b * x[mask_middle]
            result[mask_lower] = 0
            
            # Copy back to original tensor
            x.copy_(result)
            return x
        else:
            # Initialize result tensor
            result = torch.zeros_like(x)
            
            # Apply different functions to different regions
            result[mask_upper] = x[mask_upper]
            result[mask_middle] = self.a * x[mask_middle]**2 + self.b * x[mask_middle]
            # result[mask_lower] is already 0
            
            return result


@register_activation
class SinLU(nn.Module):
    r"""
    Applies the Sinu-Sigmoidal Linear Unit function:

    :math:`\text{SinLU}(x) = (x + a \cdot \sin(b \cdot x)) \cdot \sigma(x)`

    where :math:`\sigma(x)` is the sigmoid function.

    Args:
        a (float, optional): Amplitude parameter for sine component. Default: 0.5
        b (float, optional): Frequency parameter for sine component. Default: 1.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.SinLU(a=0.5, b=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.SinLU(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 0.5, 
        b: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        modified_x = x + self.a * torch.sin(self.b * x)
        result = modified_x * torch.sigmoid(x)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class ErfAct(nn.Module):
    r"""
    Applies the ErfAct activation function:

    :math:`\text{ErfAct}(x) = x \cdot \text{erf}(a \cdot \exp(b \cdot x))`

    where :math:`\text{erf}(x)` is the error function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 1.0
        b (float, optional): Parameter controlling the exponential growth. Default: 0.5
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.ErfAct(a=1.0, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.ErfAct(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 0.5, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        # Calculate exp(b*x) with clipping to prevent overflow
        exp_term = torch.exp(torch.clamp(self.b * x, max=20))
        result = x * torch.erf(self.a * exp_term)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class PSerf(nn.Module):
    r"""
    Applies the Parametric Serf activation function:

    :math:`\text{pserf}(x) = x \cdot \text{erf}(a \cdot \ln(1 + \exp(b \cdot x)))`

    where :math:`\text{erf}(x)` is the error function.

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 1.0
        b (float, optional): Parameter controlling the softplus term. Default: 1.0
        learnable (bool, optional): optionally make parameters trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.PSerf(a=1.0, b=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.PSerf(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 1.0, 
        b: float = 1.0, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = Tensor([a])
            self.b = Tensor([b])

    def forward(self, x) -> Tensor:
        # Calculate softplus: ln(1 + exp(b*x))
        softplus = torch.log(1 + torch.exp(torch.clamp(self.b * x, max=20)))
        result = x * torch.erf(self.a * softplus)
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


@register_activation
class Swim(nn.Module):
    r"""
    Applies the Swim activation function:

    :math:`\text{Swim}(x) = x \cdot \frac{1}{2} \left(1 + \frac{a \cdot x}{\sqrt{1 + x^2}}\right)`

    Args:
        a (float, optional): Parameter controlling the shape of the function. Default: 0.5
        learnable (bool, optional): optionally make ``a`` trainable. Default: ``False``
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.Swim(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.Swim(learnable=True)
        >>> x = torch.randn(2, 3, 4)
        >>> output = m(x)
    """

    def __init__(
        self, 
        a: float = 0.5, 
        learnable: bool = False, 
        inplace: bool = False
    ):
        super().__init__()
        self.inplace = inplace
        if learnable:
            self.a = nn.Parameter(Tensor([a]))
        else:
            self.a = Tensor([a])

    def forward(self, x) -> Tensor:
        # Calculate the modified sigmoid-like term
        sigmoid_term = 0.5 * (1 + (self.a * x) / torch.sqrt(1 + x.pow(2)))
        result = x * sigmoid_term
        
        if self.inplace and hasattr(x, 'copy_'):
            x.copy_(result)
            return x
        else:
            return result


import torch
import torch.nn as nn
import torch.nn.functional as F
import math

from torch import Tensor
from torch_activation import register_activation


@register_activation
class CoLU(nn.Module):
    r"""
    Applies the Collapsing Linear Unit activation function:

    :math:`\text{CoLU}(x) = \frac{x}{1-x \cdot e^{-(x + e^x)}}`

     See: https://doi.org/10.48550/arXiv.2112.12078

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/CoLU.png

    Examples::

        >>> m = nn.CoLU()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = nn.CoLU(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace=False):
        super(CoLU, self).__init__()
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        if self.inplace:
            return x.div_(1 - x * torch.exp(-1 * (x + torch.exp(x))))
        else:
            return x / (1 - x * torch.exp(-1 * (x + torch.exp(x))))


@register_activation
class Phish(torch.nn.Module):
    r"""
    Applies the Phish activation function:

    :math:`\text{Phish}(x) = x \cdot \tanh (\text{GELU} (x))`

     See: `Phish: A Novel Hyper-Optimizable Activation Function`_.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/Phish.png

    Examples:
        >>> m = Phish()
        >>> x = torch.randn(2, 3)
        >>> output = m(x)

    .. _`Phish: A Novel Hyper-Optimizable Activation Function`:
        https://www.semanticscholar.org/paper/Phish%3A-A-Novel-Hyper-Optimizable-Activation-Naveen/43eb5e22da6092d28f0e842fec53ec1a76e1ba6b
    """

    def __init__(self):
        super(Phish, self).__init__()

    def forward(self, x) -> Tensor:
        output = F.gelu(x)
        output = F.tanh(output)
        output = x * output
        return output


@register_activation
class SinLU(nn.Module):
    r"""
    Applies the Sinu-sigmoidal Linear Unit activation function:

    :math:`\text{SinLU}(x) = (x + a \cdot \sin (b \cdot x)) \sigma (x)`

     See: https://doi.org/10.3390/math10030337

    Args:
        a (float, optional): Initial value for sine function magnitude. Default: 1.0.
        b (float, optional): Initial value for sine function period. Default: 1.0.
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/SinLU.png

    Examples::

        >>> m = nn.SinLU(a=5.0, b=6.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0, inplace: bool = False):
        super(SinLU, self).__init__()
        self.alpha = nn.Parameter(Tensor([a]))
        self.beta = nn.Parameter(Tensor([b]))
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        return self._forward_inplace(x) if self.inplace else self._forward(x)

    def _forward(self, x):
        result = x + self.alpha * torch.sin(self.beta * x)
        result *= torch.sigmoid(x)
        return result

    def _forward_inplace(self, x):
        s_x = torch.sigmoid(x)
        x.add_(self.alpha * torch.sin(self.beta * x))
        x.mul_(s_x)
        return x


@register_activation
class GaussianErrorLinearUnit(nn.Module):
    r"""
    Applies the Gaussian Error Linear Unit activation function:

    :math:`\text{GELU}(z) = z \cdot \Phi(z) = z \cdot \frac{1}{2} \left( 1 + \text{erf}\left(\frac{z}{\sqrt{2}}\right) \right)`

    This is a wrapper around PyTorch's native F.gelu implementation.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = GaussianErrorLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(GaussianErrorLinearUnit, self).__init__()

    def forward(self, x) -> Tensor:
        return F.gelu(x)


@register_activation
class SymmetricalGaussianErrorLinearUnit(nn.Module):
    r"""
    Applies the Symmetrical Gaussian Error Linear Unit activation function:

    :math:`\text{SGELU}(z) = a \cdot z \cdot \text{erf}\left(\frac{z}{\sqrt{2}}\right)`

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SymmetricalGaussianErrorLinearUnit(a=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(SymmetricalGaussianErrorLinearUnit, self).__init__()
        self.a = nn.Parameter(Tensor([a]))

    def forward(self, x) -> Tensor:
        return self.a * x * torch.erf(x / math.sqrt(2))


@register_activation
class CauchyLinearUnit(nn.Module):
    r"""
    Applies the Cauchy Linear Unit activation function:

    :math:`\text{CaLU}(z) = z \cdot \Phi_{\text{Cauchy}}(z) = z \cdot \left( \frac{\arctan(z)}{\pi} + \frac{1}{2} \right)`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = CauchyLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(CauchyLinearUnit, self).__init__()

    def forward(self, x) -> Tensor:
        return x * (torch.arctan(x) / math.pi + 0.5)


@register_activation
class LaplaceLinearUnit(nn.Module):
    r"""
    Applies the Laplace Linear Unit activation function:

    :math:`\text{LaLU}(z) = z \cdot \Phi_{\text{Laplace}}(z) = z \cdot \begin{cases} 
    1 - \frac{1}{2} \exp(-z), & z \geq 0 \\ 
    \frac{1}{2} \exp(z), & z < 0 
    \end{cases}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LaplaceLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(LaplaceLinearUnit, self).__init__()

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        result[pos_mask] = x[pos_mask] * (1 - 0.5 * torch.exp(-x[pos_mask]))
        result[neg_mask] = x[neg_mask] * (0.5 * torch.exp(x[neg_mask]))
        
        return result


@register_activation
class CollapsingLinearUnit(nn.Module):
    r"""
    Applies the Collapsing Linear Unit activation function:

    :math:`\text{CoLU}(z) = z \cdot \frac{1}{1 - z \exp(-(z + \exp(z)))}`

    Args:
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = CollapsingLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = CollapsingLinearUnit(inplace=True)
        >>> x = torch.randn(2)
        >>> m(x)
    """

    def __init__(self, inplace=False):
        super(CollapsingLinearUnit, self).__init__()
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        if self.inplace:
            denom = 1 - x * torch.exp(-(x + torch.exp(x)))
            return x.div_(denom)
        else:
            denom = 1 - x * torch.exp(-(x + torch.exp(x)))
            return x / denom


@register_activation
class TripleStateSwish(nn.Module):
    r"""
    Applies the Triple State Swish activation function:

    :math:`\text{TSS}(z) = z \cdot \frac{1}{1 + \exp(-z)} \left( \frac{1}{1 + \exp(-z)} + \frac{1}{1 + \exp(-z+a)} + \frac{1}{1 + \exp(-z+b)} \right)`

    Args:
        a (float, optional): First shift parameter. Default: 1.0
        b (float, optional): Second shift parameter. Default: 2.0
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TripleStateSwish(a=1.5, b=2.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 2.0, inplace: bool = False):
        super(TripleStateSwish, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        return self._forward_inplace(x) if self.inplace else self._forward(x)

    def _forward(self, x):
        sigmoid_x = torch.sigmoid(x)
        triple_term = sigmoid_x + torch.sigmoid(x - self.a) + torch.sigmoid(x - self.b)
        return x * sigmoid_x * triple_term

    def _forward_inplace(self, x):
        sigmoid_x = torch.sigmoid(x)
        triple_term = sigmoid_x + torch.sigmoid(x - self.a) + torch.sigmoid(x - self.b)
        x.mul_(sigmoid_x * triple_term)
        return x


@register_activation
class GeneralizedSwish(nn.Module):
    r"""
    Applies the Generalized Swish activation function:

    :math:`\text{GSwish}(z) = z \cdot \sigma(\exp(-z))`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = GeneralizedSwish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(GeneralizedSwish, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.sigmoid(torch.exp(-x))


@register_activation
class ExponentialSwish(nn.Module):
    r"""
    Applies the Exponential Swish activation function:

    :math:`\text{ExponentialSwish}(z) = \exp(-z) \cdot \sigma(z)`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ExponentialSwish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ExponentialSwish, self).__init__()

    def forward(self, x) -> Tensor:
        return torch.exp(-x) * torch.sigmoid(x)


@register_activation
class DerivativeOfSigmoidFunction(nn.Module):
    r"""
    Applies the Derivative of Sigmoid Function activation:

    :math:`\text{DerivativeOfSigmoidFunction}(z) = \exp(-z) \cdot (\sigma(z))^2`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = DerivativeOfSigmoidFunction()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(DerivativeOfSigmoidFunction, self).__init__()

    def forward(self, x) -> Tensor:
        sigmoid_x = torch.sigmoid(x)
        return torch.exp(-x) * sigmoid_x * sigmoid_x


@register_activation
class Gish(nn.Module):
    r"""
    Applies the Gish activation function:

    :math:`\text{Gish}(z) = z \cdot \ln(2 - \exp(-\exp(z)))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Gish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Gish, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.log(2 - torch.exp(-torch.exp(x)))


@register_activation
class Logish(nn.Module):
    r"""
    Applies the Logish activation function:

    :math:`\text{Logish}(z) = z \cdot \ln(1 + \sigma(z))`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Logish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Logish, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.log(1 + torch.sigmoid(x))


@register_activation
class LogLogish(nn.Module):
    r"""
    Applies the LogLogish activation function:

    :math:`\text{LogLogish}(z) = z \cdot (1 - \exp(-\exp(z)))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LogLogish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(LogLogish, self).__init__()

    def forward(self, x) -> Tensor:
        return x * (1 - torch.exp(-torch.exp(x)))


@register_activation
class ExpExpish(nn.Module):
    r"""
    Applies the ExpExpish activation function:

    :math:`\text{ExpExpish}(z) = z \cdot \exp(-\exp(-z))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ExpExpish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ExpExpish, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.exp(-torch.exp(-x))


@register_activation
class SelfArctan(nn.Module):
    r"""
    Applies the SelfArctan activation function:

    :math:`\text{SelfArctan}(z) = z \cdot \arctan(z)`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SelfArctan()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(SelfArctan, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.arctan(x)


@register_activation
class ParametricLogish(nn.Module):
    r"""
    Applies the Parametric Logish activation function:

    :math:`\text{ParametricLogish}(z_i) = a \cdot z_i \cdot \ln(1 + \sigma(b \cdot z_i))`

    where :math:`\sigma` is the sigmoid function.

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Sigmoid scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ParametricLogish(a=1.5, b=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0):
        super(ParametricLogish, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))

    def forward(self, x) -> Tensor:
        return self.a * x * torch.log(1 + torch.sigmoid(self.b * x))


@register_activation
class Phish(nn.Module):
    r"""
    Applies the Phish activation function:

    :math:`\text{Phish}(z) = z \cdot \tanh(\text{GELU}(z))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Phish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Phish, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.tanh(F.gelu(x))


@register_activation
class Suish(nn.Module):
    r"""
    Applies the Suish activation function:

    :math:`\text{Suish}(z) = \max(z, z \cdot \exp(-|z|))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Suish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Suish, self).__init__()

    def forward(self, x) -> Tensor:
        return torch.maximum(x, x * torch.exp(-torch.abs(x)))


@register_activation
class TangentSigmoidReLU(nn.Module):
    r"""
    Applies the Tangent Sigmoid ReLU activation function:

    :math:`\text{TangentSigmoidReLU}(z) = z \cdot \tanh(\sigma(z))`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TangentSigmoidReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(TangentSigmoidReLU, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.tanh(torch.sigmoid(x))


@register_activation
class TangentBipolarSigmoidReLU(nn.Module):
    r"""
    Applies the Tangent Bipolar Sigmoid ReLU activation function:

    :math:`\text{TangentBipolarSigmoidReLU}(z) = z \cdot \tanh\left(\frac{1 - \exp(-z)}{1 + \exp(-z)}\right)`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TangentBipolarSigmoidReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(TangentBipolarSigmoidReLU, self).__init__()

    def forward(self, x) -> Tensor:
        exp_neg_x = torch.exp(-x)
        bipolar_sigmoid = (1 - exp_neg_x) / (1 + exp_neg_x)
        return x * torch.tanh(bipolar_sigmoid)


@register_activation
class LogSigmoid(nn.Module):
    r"""
    Applies the LogSigmoid activation function:

    :math:`\text{LogSigmoid}(z) = \ln(\sigma(z)) = \ln\left(\frac{1}{1 + \exp(-z)}\right)`

    This is a wrapper around PyTorch's native F.logsigmoid implementation.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LogSigmoid()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(LogSigmoid, self).__init__()

    def forward(self, x) -> Tensor:
        return F.logsigmoid(x)


@register_activation
class DerivativeOfSiLU(nn.Module):
    r"""
    Applies the Derivative of SiLU activation function:

    :math:`\text{DerivativeOfSiLU}(z) = \sigma(z) \cdot (1 + z \cdot (1 - \sigma(z)))`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = DerivativeOfSiLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(DerivativeOfSiLU, self).__init__()

    def forward(self, x) -> Tensor:
        sigmoid_x = torch.sigmoid(x)
        return sigmoid_x * (1 + x * (1 - sigmoid_x))


@register_activation
class DoubleSiLU(nn.Module):
    r"""
    Applies the Double SiLU activation function:

    :math:`\text{DoubleSiLU}(z) = z \cdot \frac{1}{1 + \exp\left(-z \cdot \frac{1}{1 + \exp(-z)}\right)}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = DoubleSiLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(DoubleSiLU, self).__init__()

    def forward(self, x) -> Tensor:
        silu_x = x * torch.sigmoid(x)
        return x * torch.sigmoid(silu_x)


@register_activation
class ModifiedSiLU(nn.Module):
    r"""
    Applies the Modified SiLU activation function:

    :math:`\text{ModifiedSiLU}(z) = z \cdot \sigma(z) + \exp\left(-\frac{z^2}{4}\right)`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ModifiedSiLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ModifiedSiLU, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.sigmoid(x) + torch.exp(-x.pow(2) / 4)


@register_activation
class HyperbolicTangentSiLU(nn.Module):
    r"""
    Applies the Hyperbolic Tangent SiLU activation function:

    :math:`\text{HyperbolicTangentSiLU}(z) = \frac{\exp\left(\frac{z}{1 + \exp(-z)}\right) - \exp\left(-\frac{z}{1 + \exp(-z)}\right)}{\exp\left(\frac{z}{1 + \exp(-z)}\right) + \exp\left(\frac{z}{1 + \exp(-z)}\right)}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = HyperbolicTangentSiLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(HyperbolicTangentSiLU, self).__init__()

    def forward(self, x) -> Tensor:
        silu_x = x * torch.sigmoid(x)
        return torch.tanh(silu_x)

@register_activation
class ArctanSiLU(nn.Module):
    r"""
    Applies the Arctan SiLU activation function:

    :math:`\text{ArctanSiLU}(z) = \arctan(z) \cdot \frac{1}{1 + \exp(-z)}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ArctanSiLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ArctanSiLU, self).__init__()

    def forward(self, x) -> Tensor:
        return torch.arctan(x) * torch.sigmoid(x)


@register_activation
class SwAT(nn.Module):
    r"""
    Applies the SwAT activation function:

    :math:`\text{SwAT}(z) = z \cdot \frac{1}{1 + \exp(-\arctan(|z|))}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SwAT()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(SwAT, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.sigmoid(torch.arctan(torch.abs(x)))


@register_activation
class RectifiedHyperbolicSecant(nn.Module):
    r"""
    Applies the Rectified Hyperbolic Secant activation function:

    :math:`\text{RectifiedHyperbolicSecant}(z) = z \cdot \text{sech}(z)`

    where sech is the hyperbolic secant function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = RectifiedHyperbolicSecant()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(RectifiedHyperbolicSecant, self).__init__()

    def forward(self, x) -> Tensor:
        return x * (2 / (torch.exp(x) + torch.exp(-x)))


@register_activation
class LinearlyScaledHyperbolicTangent(nn.Module):
    r"""
    Applies the Linearly Scaled Hyperbolic Tangent activation function:

    :math:`\text{LinearlyScaledHyperbolicTangent}(z) = z \cdot \tanh(z)`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LinearlyScaledHyperbolicTangent()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(LinearlyScaledHyperbolicTangent, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.tanh(x)


@register_activation
class Mish(nn.Module):
    r"""
    Applies the Mish activation function:

    :math:`\text{Mish}(z) = z \cdot \tanh(\text{softplus}(z)) = z \cdot \tanh(\ln(1 + \exp(z)))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Mish()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Mish, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.tanh(F.softplus(x))


@register_activation
class Smish(nn.Module):
    r"""
    Applies the Smish activation function:

    :math:`\text{Smish}(z) = a \cdot z \cdot \tanh(\ln(1 + \sigma(b \cdot z)))`

    where :math:`\sigma` is the sigmoid function.

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Sigmoid scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Smish(a=1.5, b=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0):
        super(Smish, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))

    def forward(self, x) -> Tensor:
        return self.a * x * torch.tanh(torch.log(1 + torch.sigmoid(self.b * x)))


@register_activation
class TanhExp(nn.Module):
    r"""
    Applies the TanhExp activation function:

    :math:`\text{TanhExp}(z) = z \cdot \tanh(\exp(z))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TanhExp()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(TanhExp, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.tanh(torch.exp(x))


@register_activation
class Serf(nn.Module):
    r"""
    Applies the Serf activation function:

    :math:`\text{Serf}(z) = z \cdot \text{erf}(\ln(1 + \exp(z)))`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = Serf()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(Serf, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.erf(F.softplus(x))


@register_activation
class EfficientAsymmetricNonlinearActivationFunction(nn.Module):
    r"""
    Applies the Efficient Asymmetric Nonlinear Activation Function:

    :math:`\text{EfficientAsymmetricNonlinearActivationFunction}(z) = z \cdot \frac{\exp(z)}{\exp(z) + 2}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = EfficientAsymmetricNonlinearActivationFunction()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(EfficientAsymmetricNonlinearActivationFunction, self).__init__()

    def forward(self, x) -> Tensor:
        exp_x = torch.exp(x)
        return x * (exp_x / (exp_x + 2))


@register_activation
class SinSig(nn.Module):
    r"""
    Applies the SinSig activation function:

    :math:`\text{SinSig}(z) = z \cdot \sin\left(\frac{\pi}{2} \sigma(z)\right)`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SinSig()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(SinSig, self).__init__()

    def forward(self, x) -> Tensor:
        return x * torch.sin((math.pi / 2) * torch.sigmoid(x))


@register_activation
class SiELU(nn.Module):
    r"""
    Applies the SiELU activation function:

    :math:`\text{SiELU}(z) = z \cdot \sigma\left(\sqrt{\frac{2}{\pi}} z + 0.044715 z^3\right)`

    where :math:`\sigma` is the sigmoid function.

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SiELU()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(SiELU, self).__init__()

    def forward(self, x) -> Tensor:
        inner = math.sqrt(2/math.pi) * x + 0.044715 * x.pow(3)
        return x * torch.sigmoid(inner)


import os
import math
import torch
import plotly
import plotly.io as pio
import plotly.graph_objects as go


def plot_activation(
    activation: torch.nn.Module,
    params: dict = {},
    save_dir="./images/activation_images",
    x_range=(-5, 5),
    y_range=None,
    preview=False,
    plot_derivative=True,
):
    """
    Plot the activation function and optionally its derivative.

    Parameters:
        activation (torch.nn.Module): The activation function to plot.
        params (dict): A dictionary of parameter names and values for the activation function.
        save_dir (str, optional): The directory to save the generated image. Defaults to "./images/activation_images".
        x_range (tuple, optional): The x-axis range for the plot. Defaults to (-5, 5).
        y_range (tuple, optional): The y-axis range for the plot. Defaults to None (auto-scale).
        preview (bool, optional): Whether to display the plot interactively. Defaults to False.
        plot_derivative (bool, optional): Whether to plot the derivative of the activation function. Defaults to True.

    Returns:
        None

    The function plots the activation function and, optionally, its derivative for the given parameters.
    The resulting plot is saved as an image in the specified `save_dir` directory.

    If `preview` is set to True, the plot will also be displayed interactively.

    Example::

        >>> # Plotting the sigmoid activation function and its derivative
        >>> params = {'n': 1.0}
        >>> plot_activation(torch.nn.Sigmoid(), params)

    """
    x = torch.linspace(x_range[0], x_range[1], 1000)
    x.requires_grad = True  # Enable gradient computation for x

    fig = go.Figure()

    num_plots = max(len(v) for v in params.values()) if params else 0

    # Color for each param
    colors = plotly.colors.qualitative.D3[: max(1, num_plots)]

    if num_plots == 0:
        # Add default plot with no parameter variations
        m = activation()
        y = m(x)

        if y_range is None:
            y_min = torch.min(y).item()
            y_max = torch.max(y).item()
            y_padding = (y_max - y_min) * 0.1  # Add a 10% padding
            y_range = (y_min - y_padding, y_max + y_padding)

        fig.add_trace(
            go.Scatter(
                x=x.detach().numpy(),
                y=y.detach().numpy(),
                name=activation.__name__,
                line=dict(color=colors[0]),
            )
        )

        if plot_derivative:
            d = torch.autograd.grad(y, x, torch.ones_like(y), create_graph=True)[0]
            fig.add_trace(
                go.Scatter(
                    x=x.detach().numpy(),
                    y=d.detach().numpy(),
                    name=f"d/dx {activation.__name__}",
                    line=dict(color=colors[0], dash="dot"),
                )
            )

    else:
        param_combinations = torch.tensor(
            [[v for v in params[key]] for key in params.keys()]
        ).T

        y_ = []
        for i, combination in enumerate(param_combinations):
            kwargs = {key: value for key, value in zip(params.keys(), combination)}

            m = activation(**kwargs)
            y = m(x)
            y_.append(y)

            label = (
                f"{activation.__name__}("
                + ", ".join([f"{key}={value}" for key, value in kwargs.items()])
                + ")"
            )

            fig.add_trace(
                go.Scatter(
                    x=x.detach().numpy(),
                    y=y.detach().numpy(),
                    name=label,
                    line=dict(color=colors[i % len(colors)]),
                )
            )

            if plot_derivative:
                d = torch.autograd.grad(y, x, torch.ones_like(y), create_graph=True)[0]
                fig.add_trace(
                    go.Scatter(
                        x=x.detach().numpy(),
                        y=d.detach().numpy(),
                        name=f"d/dx {label}",
                        line=dict(color=colors[i % len(colors)], dash="dot"),
                    )
                )

        y_ = torch.stack(y_)

        if y_range is None:
            y_min = torch.min(y_).item()
            y_max = torch.max(y_).item()
            y_padding = (y_max - y_min) * 0.1  # Add a 10% padding
            y_range = (y_min - y_padding, y_max + y_padding)

    y_range = list(y_range)
    if y_range[1] - y_range[0] > 3:
        y_range[0] = math.floor(y_range[0])
        y_range[1] = math.ceil(y_range[1])

    fig.update_layout(
        paper_bgcolor="rgba(0,0,0,0)",
        xaxis=dict(range=x_range),
        yaxis=dict(range=y_range),
        legend=dict(orientation="h", yanchor="top", y=-0.2, xanchor="center", x=0.5),
        shapes=[
            dict(
                type="line",
                x0=0,
                x1=0,
                y0=y_range[0],
                y1=y_range[1],
                line=dict(color="black", width=1),
            ),
            dict(
                type="line",
                x0=x_range[0],
                x1=x_range[1],
                y0=0,
                y1=0,
                line=dict(color="black", width=1),
            ),
        ],
        xaxis_tickvals=list(range(math.floor(x_range[0]), math.ceil(x_range[1]) + 1)),
        yaxis_tickvals=list(range(math.floor(y_range[0]), math.ceil(y_range[1]) + 1)),
    )

    if preview:
        fig.show()

    os.makedirs(save_dir, exist_ok=True)
    file_name = os.path.join(save_dir, f"{activation.__name__}.png")
    pio.write_image(fig, file_name)
    print(f"Image saved as {file_name}")


if __name__ == "__main__":
    from relu import SoftsignRReLU

    plot_activation(SoftsignRReLU, preview=True)



import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from torch import Tensor
from scipy.special import bernoulli

from torch_activation import register_activation

# TODO: Naming

class FAAF(nn.Module):
    r"""
    Applies the Fractional Adaptive Activation Function (FAAF):

    :math:`\text{FAAF}(z) = D^a f(z)`

    where :math:`D^a` is the a-th fractional derivative of function f.

    This is a base class for fractional activation functions.

    Args:
        a_init (float, optional): Initial value for the fractional order parameter a. Default: 0.5

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.
    """

    def __init__(self, a_init: float = 0.5):
        super(FAAF, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))

    def forward(self, x) -> Tensor:
        raise NotImplementedError("Subclasses must implement forward method")


@register_activation
class FracReLU(nn.Module):
    r"""
    Applies the Fractional ReLU function:

    :math:`\text{FracReLU}(z_i) = \frac{z_i^{1 - a_i}}{\Gamma(2 - a_i)}`

    where :math:`\Gamma` is the Gamma function and :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracReLU(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, eps: float = 1e-6):
        super(FracReLU, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.eps = eps

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # Only apply to positive values (like ReLU)
        pos_mask = x > 0
        result = torch.zeros_like(x)
        
        # Apply fractional operation only to positive values
        if pos_mask.any():
            x_pos = x[pos_mask]
            # Add eps for numerical stability when x is close to zero
            x_pos = x_pos + self.eps
            gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
            result[pos_mask] = torch.pow(x_pos, 1 - a_clamped) / gamma_term
            
        return result


@register_activation
class FracSoftplus(nn.Module):
    r"""
    Applies the Fractional Softplus function:

    :math:`\text{FracSoftplus}(z_i) = D^{a_i} \ln(1 + \exp(z_i))`

    Approximated using a finite sum:

    :math:`\text{FracSoftplus}(z_i) \approx \frac{1}{h^{a_i}} \sum_{n=0}^{N} \frac{(-1)^n \Gamma(a_i + 1) \ln(1 + \exp(z_i - nh))}{\Gamma(n + 1) \Gamma(1 - n + a_i)}`

    where :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        h (float, optional): Step size for approximation. Default: 0.1
        n_terms (int, optional): Number of terms in the approximation sum. Default: 5

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracSoftplus(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, h: float = 0.1, n_terms: int = 5):
        super(FracSoftplus, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.h = h
        self.n_terms = n_terms

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 2) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 1.99)
        
        # Special cases for efficiency
        if torch.abs(a_clamped - 0.0) < 1e-4:
            # When a ≈ 0, return softplus
            return F.softplus(x)
        elif torch.abs(a_clamped - 1.0) < 1e-4:
            # When a ≈ 1, return sigmoid
            return torch.sigmoid(x)
        
        # Compute the approximation using finite sum
        result = torch.zeros_like(x)
        h_pow_a = torch.pow(self.h, a_clamped)
        gamma_a_plus_1 = torch.exp(torch.lgamma(a_clamped + 1))
        
        for n in range(self.n_terms):
            # Calculate terms in the sum
            coef = ((-1) ** n) * gamma_a_plus_1
            denom = math.factorial(n) * torch.exp(torch.lgamma(1 - n + a_clamped))
            term = coef * F.softplus(x - n * self.h) / denom
            result += term
            
        return result / h_pow_a


@register_activation
class FracTanh(nn.Module):
    r"""
    Applies the Fractional Hyperbolic Tangent function:

    :math:`\text{FracTanh}(z_i) = D^{a_i} \tanh(z_i)`

    Approximated using a finite sum:

    :math:`\text{FracTanh}(z_i) \approx \frac{1}{h^{a_i}} \sum_{n=0}^{N} \frac{(-1)^n \Gamma(a_i + 1) \tanh(z_i - nh)}{\Gamma(n + 1) \Gamma(1 - n + a_i)}`

    where :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        h (float, optional): Step size for approximation. Default: 0.1
        n_terms (int, optional): Number of terms in the approximation sum. Default: 5

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracTanh(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, h: float = 0.1, n_terms: int = 5):
        super(FracTanh, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.h = h
        self.n_terms = n_terms

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 2) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 1.99)
        
        # Special cases for efficiency
        if torch.abs(a_clamped - 0.0) < 1e-4:
            # When a ≈ 0, return tanh
            return torch.tanh(x)
        
        # Compute the approximation using finite sum
        result = torch.zeros_like(x)
        h_pow_a = torch.pow(self.h, a_clamped)
        gamma_a_plus_1 = torch.exp(torch.lgamma(a_clamped + 1))
        
        for n in range(self.n_terms):
            # Calculate terms in the sum
            coef = ((-1) ** n) * gamma_a_plus_1
            denom = math.factorial(n) * torch.exp(torch.lgamma(1 - n + a_clamped))
            term = coef * torch.tanh(x - n * self.h) / denom
            result += term
            
        return result / h_pow_a


@register_activation
class FALU(nn.Module):
    r"""
    Applies the Fractional Adaptive Linear Unit (FALU):

    :math:`\text{FALU}(z_i) = D^{a_i} z_i \sigma(b_i z_i)`

    where :math:`\sigma` is the sigmoid function, and :math:`a_i` and :math:`b_i` are trainable parameters.

    For computational efficiency, an approximation is used:

    :math:`\text{FALU}(z_i) \approx \begin{cases} 
    g(z_i, b_i) + a_i \sigma(b_i z_i) (1 - g(z_i, b_i)), & a_i \in [0, 1] \\
    g(z_i, b_i) + a_i \sigma(b_i z_i) (1 - 2h(z_i, b_i)), & a_i \in (1, 2] 
    \end{cases}`

    where:
    :math:`g(z_i, b_i) = z_i \sigma(b_i z_i)`
    :math:`h(z_i, b_i) = g(z_i, b_i) + \sigma(z_i) (1 - g(z_i, b_i))`

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        b_init (float, optional): Initial value for the trainable parameter b. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FALU(a_init=0.3, b_init=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, b_init: float = 1.0):
        super(FALU, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.b = nn.Parameter(Tensor([b_init]))

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 2) range and b to be in (1, 10) range for stability
        a_clamped = torch.clamp(self.a, 0.0, 2.0)
        b_clamped = torch.clamp(self.b, 1.0, 10.0)
        
        # Calculate g(z, b) = z * sigmoid(b*z)
        g_z_b = x * torch.sigmoid(b_clamped * x)
        
        # Calculate h(z, b) = g(z, b) + sigmoid(z) * (1 - g(z, b))
        h_z_b = g_z_b + torch.sigmoid(x) * (1 - g_z_b)
        
        # Apply the approximation based on the value of a
        result = torch.zeros_like(x)
        
        # For a in [0, 1]
        mask_a_0_1 = (a_clamped >= 0) & (a_clamped <= 1)
        if mask_a_0_1:
            result = g_z_b + a_clamped * torch.sigmoid(b_clamped * x) * (1 - g_z_b)
        
        # For a in (1, 2]
        mask_a_1_2 = (a_clamped > 1) & (a_clamped <= 2)
        if mask_a_1_2:
            result = g_z_b + a_clamped * torch.sigmoid(b_clamped * x) * (1 - 2 * h_z_b)
            
        return result


@register_activation
class FracLReLU(nn.Module):
    r"""
    Applies the Fractional Leaky ReLU function:

    :math:`\text{FracLReLU}(z_i) = \begin{cases} 
    \frac{z_i^{1 - a_i}}{\Gamma(2 - a_i)}, & z_i \geq 0 \\
    \frac{0.1 \cdot z_i^{1 - a_i}}{\Gamma(2 - a_i)}, & z_i < 0 
    \end{cases}`

    where :math:`\Gamma` is the Gamma function and :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        negative_slope (float, optional): Controls the angle of the negative slope. Default: 0.1
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracLReLU(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, negative_slope: float = 0.1, eps: float = 1e-6):
        super(FracLReLU, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.negative_slope = negative_slope
        self.eps = eps

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # Calculate gamma term
        gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
        
        # Create masks for positive and negative values
        pos_mask = x >= 0
        neg_mask = x < 0
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply to positive values
        if pos_mask.any():
            x_pos = x[pos_mask]
            # Add eps for numerical stability when x is close to zero
            x_pos = x_pos + self.eps
            result[pos_mask] = torch.pow(x_pos, 1 - a_clamped) / gamma_term
        
        # Apply to negative values
        if neg_mask.any():
            x_neg = -x[neg_mask]  # Make positive for power operation
            # Add eps for numerical stability when x is close to zero
            x_neg = x_neg + self.eps
            result[neg_mask] = -self.negative_slope * torch.pow(x_neg, 1 - a_clamped) / gamma_term
            
        return result


@register_activation
class FracPReLU(nn.Module):
    r"""
    Applies the Fractional Parametric ReLU function:

    :math:`\text{FracPReLU}(z_i) = \begin{cases} 
    \frac{z_i^{1 - a_i}}{\Gamma(2 - a_i)}, & z_i \geq 0 \\
    \frac{b_i \cdot z_i^{1 - a_i}}{\Gamma(2 - a_i)}, & z_i < 0 
    \end{cases}`

    where :math:`\Gamma` is the Gamma function, :math:`a_i` is a fixed parameter, and :math:`b_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the parameter a. Default: 0.5
        b_init (float, optional): Initial value for the trainable parameter b. Default: 0.25
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracPReLU(a_init=0.3, b_init=0.2)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, b_init: float = 0.25, eps: float = 1e-6):
        super(FracPReLU, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.b = nn.Parameter(Tensor([b_init]))
        self.eps = eps

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # Calculate gamma term
        gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
        
        # Create masks for positive and negative values
        pos_mask = x >= 0
        neg_mask = x < 0
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply to positive values
        if pos_mask.any():
            x_pos = x[pos_mask]
            # Add eps for numerical stability when x is close to zero
            x_pos = x_pos + self.eps
            result[pos_mask] = torch.pow(x_pos, 1 - a_clamped) / gamma_term
        
        # Apply to negative values
        if neg_mask.any():
            x_neg = -x[neg_mask]  # Make positive for power operation
            # Add eps for numerical stability when x is close to zero
            x_neg = x_neg + self.eps
            # Use absolute value of b for stability
            b_abs = torch.abs(self.b)
            result[neg_mask] = -b_abs * torch.pow(x_neg, 1 - a_clamped) / gamma_term
            
        return result


@register_activation
class FracELU(nn.Module):
    r"""
    Applies the Fractional Exponential Linear Unit function:

    :math:`\text{FracELU}(z_i) = \begin{cases} 
    \frac{z_i^{1 - a_i}}{\Gamma(2 - a_i)}, & z_i \geq 0 \\
    b \sum_{k=0}^{N} \frac{1}{\Gamma(k + 1)} \frac{\Gamma(k + 1 - a_i)}{\Gamma(k + 1)} z_i^{k - a_i} - \frac{b}{\Gamma(1 - a_i)} z_i^{-a_i}, & z_i < 0 
    \end{cases}`

    where :math:`\Gamma` is the Gamma function, :math:`a_i` is a trainable parameter, and :math:`b` is a fixed parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        alpha (float, optional): The alpha parameter for ELU. Default: 1.0
        n_terms (int, optional): Number of terms in the approximation sum. Default: 5
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracELU(a_init=0.3, alpha=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, alpha: float = 1.0, n_terms: int = 5, eps: float = 1e-6):
        super(FracELU, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.alpha = alpha
        self.n_terms = n_terms
        self.eps = eps

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # Create masks for positive and negative values
        pos_mask = x >= 0
        neg_mask = x < 0
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply to positive values
        if pos_mask.any():
            x_pos = x[pos_mask]
            # Add eps for numerical stability when x is close to zero
            x_pos = x_pos + self.eps
            gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
            result[pos_mask] = torch.pow(x_pos, 1 - a_clamped) / gamma_term
        
        # Apply to negative values
        if neg_mask.any():
            x_neg = x[neg_mask]
            
            # First term: -b/Gamma(1-a) * x^(-a)
            gamma_term_1 = torch.exp(torch.lgamma(1 - a_clamped))
            first_term = -self.alpha / gamma_term_1 * torch.pow(torch.abs(x_neg) + self.eps, -a_clamped)
            
            # Second term: b * sum(...)
            second_term = torch.zeros_like(x_neg)
            for k in range(self.n_terms):
                gamma_k_plus_1 = math.factorial(k)
                gamma_k_plus_1_minus_a = torch.exp(torch.lgamma(k + 1 - a_clamped))
                coef = 1 / gamma_k_plus_1 * gamma_k_plus_1_minus_a / gamma_k_plus_1
                term = coef * torch.pow(torch.abs(x_neg) + self.eps, k - a_clamped)
                second_term += term
            
            result[neg_mask] = self.alpha * second_term + first_term
            
        return result


@register_activation
class FracSiLU1(nn.Module):
    r"""
    Applies the Fractional SiLU Variant 1 function:

    :math:`\text{FracSiLU1}(z_i) = \begin{cases} 
    \frac{z_i^{1 - a_i}}{\Gamma(2 - a_i)}, & z_i \geq 0 \\
    \sum_{k=0}^{N} \frac{(-1)^k + (2^{k+1} - 1) B_{k+1} \Gamma(k + 2)}{\Gamma(k + 2 - a_i) (k + 1)!} z_i^{k+1 - a_i}, & z_i < 0 
    \end{cases}`

    where :math:`\Gamma` is the Gamma function, :math:`B_n` is the n-th Bernoulli number, and :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        n_terms (int, optional): Number of terms in the approximation sum. Default: 5
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracSiLU1(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, n_terms: int = 5, eps: float = 1e-6):
        super(FracSiLU1, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.n_terms = n_terms
        self.eps = eps
        # Pre-compute Bernoulli numbers
        self.bernoulli_numbers = [bernoulli(i) for i in range(n_terms + 2)]

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # Create masks for positive and negative values
        pos_mask = x >= 0
        neg_mask = x < 0
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply to positive values
        if pos_mask.any():
            x_pos = x[pos_mask]
            # Add eps for numerical stability when x is close to zero
            x_pos = x_pos + self.eps
            gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
            result[pos_mask] = torch.pow(x_pos, 1 - a_clamped) / gamma_term
        
        # Apply to negative values
        if neg_mask.any():
            x_neg = x[neg_mask]
            
            # Compute the sum
            sum_term = torch.zeros_like(x_neg)
            for k in range(self.n_terms):
                # Calculate (-1)^k + (2^(k+1) - 1) * B_{k+1}
                bernoulli_term = self.bernoulli_numbers[k + 1]
                coef_term = ((-1) ** k) + ((2 ** (k + 1)) - 1) * bernoulli_term
                
                # Calculate Gamma(k + 2) / (Gamma(k + 2 - a) * (k + 1)!)
                gamma_k_plus_2 = math.factorial(k + 1)
                gamma_k_plus_2_minus_a = torch.exp(torch.lgamma(k + 2 - a_clamped))
                k_plus_1_factorial = math.factorial(k + 1)
                
                coef = coef_term * gamma_k_plus_2 / (gamma_k_plus_2_minus_a * k_plus_1_factorial)
                
                # Calculate z^(k+1-a)
                power_term = torch.pow(torch.abs(x_neg) + self.eps, k + 1 - a_clamped)
                
                sum_term += coef * power_term
            
            result[neg_mask] = sum_term
            
        return result

# TODO: Forward pass failed: Can't call numpy() on Tensor that requires grad. Use tensor.detach().numpy() instead.
# @register_activation
class FracSiLU2(nn.Module):
    r"""
    Applies the Fractional SiLU Variant 2 function:

    :math:`\text{FracSiLU2}(z_i) = \sum_{k=0}^{N} \frac{(-1)^k + (2^{k+1} - 1) B_{k+1} \Gamma(k + 2)}{\Gamma(k + 2 - a_i) (k + 1)!} z_i^{k+1 - a_i}`

    where :math:`\Gamma` is the Gamma function, :math:`B_n` is the n-th Bernoulli number, and :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        n_terms (int, optional): Number of terms in the approximation sum. Default: 5
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracSiLU2(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, n_terms: int = 5, eps: float = 1e-6):
        super(FracSiLU2, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.n_terms = n_terms
        self.eps = eps
        # Pre-compute Bernoulli numbers
        self.bernoulli_numbers = [bernoulli(i) for i in range(n_terms + 2)]

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # Compute the sum for all values
        result = torch.zeros_like(x)
        
        for k in range(self.n_terms):
            # Calculate (-1)^k + (2^(k+1) - 1) * B_{k+1}
            bernoulli_term = self.bernoulli_numbers[k + 1]
            coef_term = ((-1) ** k) + ((2 ** (k + 1)) - 1) * bernoulli_term
            
            # Calculate Gamma(k + 2) / (Gamma(k + 2 - a) * (k + 1)!)
            gamma_k_plus_2 = math.factorial(k + 1)
            gamma_k_plus_2_minus_a = torch.exp(torch.lgamma(k + 2 - a_clamped))
            k_plus_1_factorial = math.factorial(k + 1)
            
            coef = coef_term * gamma_k_plus_2 / (gamma_k_plus_2_minus_a * k_plus_1_factorial)
            
            # Calculate z^(k+1-a)
            power_term = torch.pow(torch.abs(x) + self.eps, k + 1 - a_clamped) * torch.sign(x)
            
            result += coef * power_term
        
        return result


@register_activation
class FracGELU1(nn.Module):
    r"""
    Applies the Fractional GELU Variant 1 function:

    :math:`\text{FracGELU1}(z_i) = \begin{cases} 
    \frac{z_i^{1 - a_i}}{\Gamma(2 - a_i)}, & z_i \geq 0 \\
    \frac{0.5 z_i^{1 - a_i}}{\Gamma(2 - a_i)} - \frac{1}{\sqrt{2\pi}} \sum_{k=0}^{N} \frac{1}{k!} \left(-\frac{1}{2}\right)^k \frac{\Gamma(2k + 3)}{\Gamma(2k + 3 - a_i)} z_i^{2k+1 - a_i}, & z_i < 0 
    \end{cases}`

    where :math:`\Gamma` is the Gamma function and :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        n_terms (int, optional): Number of terms in the approximation sum. Default: 5
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracGELU1(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, n_terms: int = 5, eps: float = 1e-6):
        super(FracGELU1, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.n_terms = n_terms
        self.eps = eps
        self.sqrt_2pi = math.sqrt(2 * math.pi)

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # Create masks for positive and negative values
        pos_mask = x >= 0
        neg_mask = x < 0
        
        # Initialize result tensor
        result = torch.zeros_like(x)
        
        # Apply to positive values
        if pos_mask.any():
            x_pos = x[pos_mask]
            # Add eps for numerical stability when x is close to zero
            x_pos = x_pos + self.eps
            gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
            result[pos_mask] = torch.pow(x_pos, 1 - a_clamped) / gamma_term
        
        # Apply to negative values
        if neg_mask.any():
            x_neg = x[neg_mask]
            
            # First term: 0.5 * z^(1-a) / Gamma(2-a)
            gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
            first_term = 0.5 * torch.pow(torch.abs(x_neg) + self.eps, 1 - a_clamped) / gamma_term
            
            # Second term: -1/sqrt(2π) * sum(...)
            sum_term = torch.zeros_like(x_neg)
            for k in range(self.n_terms):
                # Calculate 1/k! * (-1/2)^k
                k_factorial = math.factorial(k)
                neg_half_pow_k = ((-0.5) ** k)
                
                # Calculate Gamma(2k+3) / Gamma(2k+3-a)
                gamma_2k_plus_3 = math.factorial(2*k + 2)
                gamma_2k_plus_3_minus_a = torch.exp(torch.lgamma(2*k + 3 - a_clamped))
                
                # Calculate z^(2k+1-a)
                power_term = torch.pow(torch.abs(x_neg) + self.eps, 2*k + 1 - a_clamped)
                
                term = (1 / k_factorial) * neg_half_pow_k * (gamma_2k_plus_3 / gamma_2k_plus_3_minus_a) * power_term
                sum_term += term
            
            second_term = -(1 / self.sqrt_2pi) * sum_term
            
            result[neg_mask] = first_term + second_term
            
        return result


@register_activation
class FracGELU2(nn.Module):
    r"""
    Applies the Fractional GELU Variant 2 function:

    :math:`\text{FracGELU2}(z_i) = \frac{0.5 z_i^{1 - a_i}}{\Gamma(2 - a_i)} - \frac{1}{\sqrt{2\pi}} \sum_{k=0}^{N} \frac{1}{k!} \left(-\frac{1}{2}\right)^k \frac{\Gamma(2k + 3)}{\Gamma(2k + 3 - a_i)} z_i^{2k+1 - a_i}`

    where :math:`\Gamma` is the Gamma function and :math:`a_i` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 0.5
        n_terms (int, optional): Number of terms in the approximation sum. Default: 5
        eps (float, optional): Small constant for numerical stability. Default: 1e-6

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FracGELU2(a_init=0.3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 0.5, n_terms: int = 5, eps: float = 1e-6):
        super(FracGELU2, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.n_terms = n_terms
        self.eps = eps
        self.sqrt_2pi = math.sqrt(2 * math.pi)

    def forward(self, x) -> Tensor:
        # Clamp a to be in (0, 1) range for stability
        a_clamped = torch.clamp(self.a, 0.01, 0.99)
        
        # First term: 0.5 * z^(1-a) / Gamma(2-a)
        gamma_term = torch.exp(torch.lgamma(2 - a_clamped))
        first_term = 0.5 * torch.pow(torch.abs(x) + self.eps, 1 - a_clamped) * torch.sign(x)
        first_term = first_term / gamma_term
        
        # Second term: -1/sqrt(2π) * sum(...)
        sum_term = torch.zeros_like(x)
        for k in range(self.n_terms):
            # Calculate 1/k! * (-1/2)^k
            k_factorial = math.factorial(k)
            neg_half_pow_k = ((-0.5) ** k)
            
            # Calculate Gamma(2k+3) / Gamma(2k+3-a)
            gamma_2k_plus_3 = math.factorial(2*k + 2)
            gamma_2k_plus_3_minus_a = torch.exp(torch.lgamma(2*k + 3 - a_clamped))
            
            # Calculate z^(2k+1-a)
            power_term = torch.pow(torch.abs(x) + self.eps, 2*k + 1 - a_clamped) * torch.sign(x)
            
            term = (1 / k_factorial) * neg_half_pow_k * (gamma_2k_plus_3 / gamma_2k_plus_3_minus_a) * power_term
            sum_term += term
        
        second_term = -(1 / self.sqrt_2pi) * sum_term
        
        return first_term + second_term
